from __future__ import annotations

import sys
import uuid
from collections import namedtuple
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from kombu.utils.encoding import str_to_bytes
from kombu.utils.json import (_register_default_types, dumps, loads,
                              register_type)

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    from backports.zoneinfo import ZoneInfo


class Custom:
    def __init__(self, data):
        self.data = data

    def __json__(self):
        return self.data


class test_JSONEncoder:
    @pytest.fixture(autouse=True)
    def reset_registered_types(self):
        _register_default_types()

    @pytest.mark.freeze_time("2015-10-21")
    def test_datetime(self):
        now = datetime.utcnow()
        now_utc = now.replace(tzinfo=ZoneInfo("UTC"))

        original = {
            'datetime': now,
            'tz': now_utc,
            'date': now.date(),
            'time': now.time(),
        }

        serialized = loads(dumps(original))

        assert serialized == original

    @given(message=st.binary())
    @settings(print_blob=True)
    def test_binary(self, message):
        serialized = loads(dumps({
            'args': (message,),
        }))
        assert serialized == {
            'args': [message],
        }

    def test_Decimal(self):
        original = {'d': Decimal('3314132.13363235235324234123213213214134')}
        serialized = loads(dumps(original))

        assert serialized == original

    def test_namedtuple(self):
        Foo = namedtuple('Foo', ['bar'])
        assert loads(dumps(Foo(123))) == [123]

    def test_UUID(self):
        constructors = [
            uuid.uuid1,
            lambda: uuid.uuid3(uuid.NAMESPACE_URL, "https://example.org"),
            uuid.uuid4,
            lambda: uuid.uuid5(uuid.NAMESPACE_URL, "https://example.org"),
            # The uuids below correspond to v6, v7 and v8 respectively and were
            # generated using the package uuid6.
            lambda: uuid.UUID("1ee0b1e6-dd55-63d2-867f-88cb9205458f"),
            lambda: uuid.UUID("0188bcbb-8475-7605-a094-fe41c58df798"),
            lambda: uuid.UUID("0188bcbb-8cb2-8bf7-b3b5-fd1faa0431bd"),
        ]
        for constructor in constructors:
            id = constructor()
            loaded_value = loads(dumps({'u': id}))
            assert loaded_value == {'u': id}
            assert loaded_value["u"].version == id.version

    def test_register_type_overrides_defaults(self):
        # This type is already registered by default, let's override it
        register_type(uuid.UUID, "uuid", lambda o: "custom", lambda o: o)
        value = uuid.uuid4()
        loaded_value = loads(dumps({'u': value}))
        assert loaded_value == {'u': "custom"}

    def test_register_type_with_new_type(self):
        # Guaranteed never before seen type
        @dataclass()
        class SomeType:
            a: int

        register_type(SomeType, "some_type", lambda o: "custom", lambda o: o)
        value = SomeType(42)
        loaded_value = loads(dumps({'u': value}))
        assert loaded_value == {'u': "custom"}

    def test_register_type_with_empty_marker(self):
        register_type(
            datetime,
            None,
            lambda o: o.isoformat(),
            lambda o: "should never be used"
        )
        now = datetime.utcnow()
        serialized_str = dumps({'now': now})
        deserialized_value = loads(serialized_str)

        assert "__type__" not in serialized_str
        assert "__value__" not in serialized_str

        # Check that there is no extra deserialization happening
        assert deserialized_value == {'now': now.isoformat()}

    def test_default(self):
        with pytest.raises(TypeError):
            dumps({'o': object()})


class test_dumps_loads:

    def test_dumps_custom_object(self):
        x = {'foo': Custom({'a': 'b'})}
        assert loads(dumps(x)) == {'foo': x['foo'].__json__()}

    def test_dumps_custom_object_no_json(self):
        x = {'foo': object()}
        with pytest.raises(TypeError):
            dumps(x)

    def test_loads_memoryview(self):
        assert loads(
            memoryview(bytearray(dumps({'x': 'z'}), encoding='utf-8'))
        ) == {'x': 'z'}

    def test_loads_bytearray(self):
        assert loads(
            bytearray(dumps({'x': 'z'}), encoding='utf-8')
        ) == {'x': 'z'}

    def test_loads_bytes(self):
        assert loads(
            str_to_bytes(dumps({'x': 'z'})),
            decode_bytes=True) == {'x': 'z'}



from __future__ import annotations

import logging
from unittest.mock import Mock, call

import pytest

from kombu.transport.native_delayed_delivery import (
    CELERY_DELAYED_DELIVERY_EXCHANGE,
    bind_queue_to_native_delayed_delivery_exchange, calculate_routing_key,
    declare_native_delayed_delivery_exchanges_and_queues, level_name)


class test_native_delayed_delivery_level_name:
    def test_level_name_with_negative_level(self):
        with pytest.raises(ValueError, match="level must be a non-negative number"):
            level_name(-1)

    def test_level_name_with_level_0(self):
        assert level_name(0) == 'celery_delayed_0'

    def test_level_name_with_level_1(self):
        assert level_name(1) == 'celery_delayed_1'


class test_declare_native_delayed_delivery_exchanges_and_queues:
    def test_invalid_queue_type(self):
        with pytest.raises(ValueError, match="queue_type must be either classic or quorum"):
            declare_native_delayed_delivery_exchanges_and_queues(Mock(), 'foo')

    def test_classic_queue_type(self):
        declare_native_delayed_delivery_exchanges_and_queues(Mock(), 'classic')

    def test_quorum_queue_type(self):
        declare_native_delayed_delivery_exchanges_and_queues(Mock(), 'quorum')


class test_bind_queue_to_native_delayed_delivery_exchange:
    def test_bind_to_direct_exchange(self, caplog):
        with caplog.at_level(logging.WARNING):
            queue_mock = Mock()
            queue_mock.bind().exchange.bind().type = 'direct'
            queue_mock.bind().exchange.bind().name = 'foo'

            bind_queue_to_native_delayed_delivery_exchange(Mock(), queue_mock)

        assert len(caplog.records) == 1

        record = caplog.records[0]
        assert (record.message == "Exchange foo is a direct exchange "
                                  "and native delayed delivery do not support direct exchanges.\n"
                                  "ETA tasks published to this exchange will "
                                  "block the worker until the ETA arrives.")

    def test_bind_to_topic_exchange(self):
        queue_mock = Mock()
        queue_mock.bind().exchange.bind().type = 'topic'
        queue_mock.bind().exchange.bind().name = 'foo'
        queue_mock.bind().routing_key = 'foo'

        bind_queue_to_native_delayed_delivery_exchange(Mock(), queue_mock)
        queue_mock.bind().exchange.bind().bind_to.assert_called_once_with(
            CELERY_DELAYED_DELIVERY_EXCHANGE,
            routing_key="#.foo"
        )
        queue_mock.bind().bind_to.assert_called_once_with(
            'foo',
            routing_key="#.foo"
        )

    def test_bind_to_topic_exchange_with_multiple_routing_keys(self):
        queue_mock = Mock()
        queue_mock.bind().exchange = None

        binding1 = Mock()
        binding1.exchange.bind().type = 'topic'
        binding1.exchange.bind().name = 'exchange1'
        binding1.routing_key = 'test1'

        binding2 = Mock()
        binding2.exchange.bind().type = 'topic'
        binding2.exchange.bind().name = 'exchange2'
        binding2.routing_key = 'test2'
        queue_mock.bind().bindings = [binding1, binding2]

        bind_queue_to_native_delayed_delivery_exchange(Mock(), queue_mock)

        queue_mock.bind().bind_to.assert_has_calls([
            call('exchange1', routing_key='#.test1'),
            call('exchange2', routing_key='#.test2')
        ], any_order=True)

        queue_mock.bind().bindings[0].exchange.bind().bind_to.assert_called_once_with(
            CELERY_DELAYED_DELIVERY_EXCHANGE,
            routing_key="#.test1"
        )
        queue_mock.bind().bindings[1].exchange.bind().bind_to.assert_called_once_with(
            CELERY_DELAYED_DELIVERY_EXCHANGE,
            routing_key="#.test2"
        )

    def test_bind_to_topic_exchange_with_wildcard_routing_key(self):
        queue_mock = Mock()
        queue_mock.bind().exchange.bind().type = 'topic'
        queue_mock.bind().exchange.bind().name = 'foo'
        queue_mock.bind().routing_key = '#.foo'

        bind_queue_to_native_delayed_delivery_exchange(Mock(), queue_mock)
        queue_mock.bind().exchange.bind().bind_to.assert_called_once_with(
            CELERY_DELAYED_DELIVERY_EXCHANGE,
            routing_key="#.foo"
        )
        queue_mock.bind().bind_to.assert_called_once_with(
            'foo',
            routing_key="#.foo"
        )


class test_calculate_routing_key:
    def test_calculate_routing_key(self):
        assert (calculate_routing_key(1, 'destination')
                == '0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.1.destination')

    def test_negative_countdown(self):
        with pytest.raises(ValueError, match="countdown must be a positive number"):
            calculate_routing_key(-1, 'foo')

    def test_zero_countdown(self):
        with pytest.raises(ValueError, match="countdown must be a positive number"):
            calculate_routing_key(0, 'foo')

    def test_empty_routing_key(self):
        with pytest.raises(ValueError, match="routing_key must be non-empty"):
            calculate_routing_key(1, '')

    def test_none_routing_key(self):
        with pytest.raises(ValueError, match="routing_key must be non-empty"):
            calculate_routing_key(1, None)



from __future__ import annotations

from kombu import Connection, Exchange, Queue
from kombu.transport.native_delayed_delivery import (
    bind_queue_to_native_delayed_delivery_exchange, calculate_routing_key,
    declare_native_delayed_delivery_exchanges_and_queues, level_name)

with Connection('amqp://guest:guest@localhost:5672//') as connection:
    declare_native_delayed_delivery_exchanges_and_queues(connection, 'quorum')
    channel = connection.channel()

    destination_exchange = Exchange('destination_exchange', type='topic')
    queue = Queue("destination", exchange=destination_exchange, routing_key='destination_route')
    queue.declare(channel=connection.channel())

    bind_queue_to_native_delayed_delivery_exchange(connection, queue)
    with connection.Producer(channel=channel) as producer:
        routing_key = calculate_routing_key(30, 'destination_route')
        producer.publish(
            "delayed msg",
            routing_key=routing_key,
            exchange=level_name(27)
        )



from __future__ import annotations

import json
from unittest import mock
from unittest.mock import MagicMock, Mock

from kombu.asynchronous.aws.ext import AWSRequest, boto3
from kombu.asynchronous.aws.sqs.connection import (AsyncSQSConnection,
                                                   _query_object_encode)
from kombu.asynchronous.aws.sqs.message import AsyncMessage
from kombu.asynchronous.aws.sqs.queue import AsyncQueue
from kombu.utils.uuid import uuid
from t.mocks import PromiseMock

from ..case import AWSCase

SQS_URL = 'https://sqs.us-west-2.amazonaws.com/'


class test_AsyncSQSConnection(AWSCase):

    def setup_method(self):
        session = boto3.session.Session(
            aws_access_key_id='AAA',
            aws_secret_access_key='AAAA',
            region_name='us-west-2',
        )
        self.sqs_client = session.client('sqs')
        self.x = AsyncSQSConnection(self.sqs_client, 'ak', 'sk', http_client=Mock())
        self.x.get_object = Mock(name='X.get_object')
        self.x.get_status = Mock(name='X.get_status')
        self.x.get_list = Mock(name='X.get_list')
        self.callback = PromiseMock(name='callback')

        self.sqs_client.get_queue_url = MagicMock(return_value={
            'QueueUrl': 'http://aws.com'
        })

    def MockRequest(self):
        return AWSRequest(
            method='POST',
            url='https://aws.com',
        )

    def MockOperationModel(self, operation_name, method):
        mock = MagicMock()
        mock.configure_mock(
            http=MagicMock(
                get=MagicMock(
                    return_value=method,
                )
            ),
            name=operation_name,
            metadata={
                'jsonVersion': '1.0',
                'targetPrefix': 'sqs',
            }
        )
        return mock

    def MockServiceModel(self, operation_name, method):
        service_model = MagicMock()
        service_model.protocol = 'json',
        service_model.operation_model = MagicMock(
            return_value=self.MockOperationModel(operation_name, method)
        )
        return service_model

    def assert_requests_equal(self, req1, req2):
        assert req1.url == req2.url
        assert req1.method == req2.method
        assert req1.data == req2.data
        assert req1.params == req2.params
        assert dict(req1.headers) == dict(req2.headers)

    def test_fetch_attributes_on_construction(self):
        """Verify default fetch_message_attributes can be set at construction."""
        x = AsyncSQSConnection(
            self.sqs_client, 'ak', 'sk', http_client=Mock(),
            fetch_message_attributes=["AttributeOne"],
        )
        assert x.fetch_message_attributes == ["AttributeOne"]

        # Default value for backwards compatibility
        assert self.x.fetch_message_attributes == ["ApproximateReceiveCount"]

    def test_create_query_request_get(self):
        # Query Protocol GET call per
        # https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-making-api-requests-xml.html
        operation_name = 'CreateQueue'
        params = {
            'DefaultVisibilityTimeout': 40,
            'QueueName': 'celery-test',
            'Version': '2012-11-05',
        }
        verb = 'GET'
        req = self.x._create_query_request(operation_name, params, SQS_URL, verb)
        self.assert_requests_equal(req, AWSRequest(
            url=SQS_URL,
            method=verb,
            data=None,
            params={
                'Action': operation_name,
                **params
            },
            headers={},
        ))

        prepared = req.prepare()  # without signing for test

        assert prepared.method == 'GET'
        assert prepared.url == (
            'https://sqs.us-west-2.amazonaws.com/?'
            'DefaultVisibilityTimeout=40'
            '&QueueName=celery-test'
            '&Version=2012-11-05'
            '&Action=CreateQueue'
        )
        assert prepared.headers == {}
        assert prepared.body is None

    def test_create_query_request(self):
        operation_name = 'ReceiveMessage'
        params = {
            'MaxNumberOfMessages': 10,
            'AttributeName.1': 'ApproximateReceiveCount',
            'WaitTimeSeconds': 20
        }
        queue_url = f'{SQS_URL}123456789012/celery-test'
        verb = 'POST'
        req = self.x._create_query_request(operation_name, params, queue_url,
                                           verb)
        self.assert_requests_equal(req, AWSRequest(
            url=queue_url,
            method=verb,
            data={
                'Action': operation_name,
                **params
            },
            headers={
                'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
            },
        ))

        prepared = req.prepare()  # without signing for test

        assert prepared.method == 'POST'
        assert prepared.url == queue_url
        assert prepared.headers == {
            'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
            'Content-Length': mock.ANY,
        }
        assert prepared.body == (
            'MaxNumberOfMessages=10'
            '&AttributeName.1=ApproximateReceiveCount'
            '&WaitTimeSeconds=20'
            '&Action=ReceiveMessage'
        )

    def test_create_json_request(self):
        operation_name = 'ReceiveMessage'
        method = 'POST'
        params = {
            'MaxNumberOfMessages': 10,
            'AttributeNames': ['ApproximateReceiveCount'],
            'WaitTimeSeconds': 20
        }
        queue_url = f'{SQS_URL}123456789012/celery-test'

        self.x.sqs_connection = Mock()
        self.x.sqs_connection._request_signer = Mock()
        self.x.sqs_connection._endpoint.host = SQS_URL
        self.x.sqs_connection.meta.service_model = Mock()
        self.x.sqs_connection.meta.service_model.protocol = 'json',
        self.x.sqs_connection.meta.service_model.operation_model = MagicMock(
            return_value=self.MockOperationModel(operation_name, method)
        )

        req = self.x._create_json_request(operation_name, params, queue_url)
        self.assert_requests_equal(req, AWSRequest(
            url=SQS_URL,
            method=method,
            data=json.dumps({
                **params,
                "QueueUrl": queue_url
            }).encode(),
            headers={
                'Content-Type': 'application/x-amz-json-1.0',
                'X-Amz-Target': f'sqs.{operation_name}'
            },
        ))

        prepared = req.prepare()  # without signing for test
        assert prepared.method == 'POST'
        assert prepared.url == SQS_URL
        assert prepared.headers == {
            'Content-Type': 'application/x-amz-json-1.0',
            'X-Amz-Target': 'sqs.ReceiveMessage',
            'Content-Length': mock.ANY,
        }
        assert json.loads(prepared.body) == {
            'MaxNumberOfMessages': 10,
            'AttributeNames': ['ApproximateReceiveCount'],
            'WaitTimeSeconds': 20,
            'QueueUrl': queue_url,
        }

    def test_make_request__with_query_protocol(self):
        # Do the necessary mocking.
        self.x.sqs_connection = Mock()
        self.x.sqs_connection._request_signer = Mock()
        self.x.sqs_connection.meta.service_model.protocol = 'query'
        self.x._create_query_request = Mock(return_value=self.MockRequest())

        # Execute the make_request called and confirm we are creating a
        # query request.
        operation = 'ReceiveMessage',
        params = {
            'MaxNumberOfMessages': 10,
            'WaitTimeSeconds': 20
        }
        pparams = {
            'json': {'AttributeNames': ['ApproximateReceiveCount']},
            'query': {'AttributeName.1': 'ApproximateReceiveCount'},
        }
        queue_url = f'{SQS_URL}123456789012/celery-test'
        verb = 'POST'

        expect_params = {**params, 'AttributeName.1': 'ApproximateReceiveCount'}

        self.x.make_request(operation, params, queue_url, verb, protocol_params=pparams)
        self.x._create_query_request.assert_called_with(
            operation, expect_params, queue_url, verb
        )

    def test_make_request__with_json_protocol(self):
        # Do the necessary mocking.
        self.x.sqs_connection = Mock()
        self.x.sqs_connection._request_signer = Mock()
        self.x.sqs_connection.meta.service_model.protocol = 'json'
        self.x._create_json_request = Mock(return_value=self.MockRequest())

        # Execute the make_request called and confirm we are creating a
        # query request.
        operation = 'ReceiveMessage',
        params = {
            'MaxNumberOfMessages': 10,
            'WaitTimeSeconds': 20
        }
        pparams = {
            'json': {'AttributeNames': ['ApproximateReceiveCount']},
            'query': {'AttributeName.1': 'ApproximateReceiveCount'},
        }

        queue_url = f'{SQS_URL}123456789012/celery-test'
        verb = 'POST'
        expect_params = {**params, 'AttributeNames': ['ApproximateReceiveCount']}

        self.x.make_request(operation, params, queue_url, verb, protocol_params=pparams)
        self.x._create_json_request.assert_called_with(
            operation, expect_params, queue_url
        )

    def test_create_queue(self):
        self.x.create_queue('foo', callback=self.callback)
        self.x.get_object.assert_called_with(
            'CreateQueue', {'QueueName': 'foo'},
            callback=self.callback,
        )

    def test_create_queue__with_visibility_timeout(self):
        self.x.create_queue(
            'foo', visibility_timeout=33, callback=self.callback,
        )
        self.x.get_object.assert_called_with(
            'CreateQueue', {
                'QueueName': 'foo',
                'DefaultVisibilityTimeout': '33'
            },
            callback=self.callback
        )

    def test_delete_queue(self):
        queue = Mock(name='queue')
        self.x.delete_queue(queue, callback=self.callback)
        self.x.get_status.assert_called_with(
            'DeleteQueue', None, queue.id, callback=self.callback,
        )

    def test_get_queue_attributes(self):
        queue = Mock(name='queue')
        self.x.get_queue_attributes(
            queue, attribute='QueueSize', callback=self.callback,
        )
        self.x.get_object.assert_called_with(
            'GetQueueAttributes', {'AttributeName': 'QueueSize'},
            queue.id, callback=self.callback,
        )

    def test_set_queue_attribute(self):
        queue = Mock(name='queue')
        self.x.set_queue_attribute(
            queue, 'Expires', '3600', callback=self.callback,
        )
        self.x.get_status.assert_called_with(
            'SetQueueAttribute',
            {}, queue.id, callback=self.callback,
            protocol_params={
                'json': {'Attributes': {'Expires': '3600'}},
                'query': {'Attribute.Name': 'Expires', 'Attribute.Value': '3600'},
            },
        )

    def test_receive_message(self):
        queue = Mock(name='queue')
        self.x.receive_message(
            queue,
            self.x.get_queue_url('queue'),
            4,
            callback=self.callback,
        )
        self.x.get_list.assert_called_with(
            'ReceiveMessage', {
                'MaxNumberOfMessages': 4,
            },
            [('Message', AsyncMessage)],
            'http://aws.com', callback=self.callback,
            parent=queue,
            protocol_params={
                'json': {'AttributeNames': ['ApproximateReceiveCount']},
                'query': {'AttributeName.1': 'ApproximateReceiveCount'},
            },
        )

    def test_receive_message__with_visibility_timeout(self):
        queue = Mock(name='queue')
        self.x.receive_message(
            queue,
            self.x.get_queue_url('queue'),
            4,
            3666,
            callback=self.callback,
        )
        self.x.get_list.assert_called_with(
            'ReceiveMessage', {
                'MaxNumberOfMessages': 4,
                'VisibilityTimeout': 3666,
            },
            [('Message', AsyncMessage)],
            'http://aws.com', callback=self.callback,
            parent=queue,
            protocol_params={
                'json': {'AttributeNames': ['ApproximateReceiveCount']},
                'query': {'AttributeName.1': 'ApproximateReceiveCount'},
            },
        )

    def test_receive_message__with_wait_time_seconds(self):
        queue = Mock(name='queue')
        self.x.receive_message(
            queue,
            self.x.get_queue_url('queue'),
            4,
            wait_time_seconds=303,
            callback=self.callback,
        )
        self.x.get_list.assert_called_with(
            'ReceiveMessage', {
                'MaxNumberOfMessages': 4,
                'WaitTimeSeconds': 303,
            },
            [('Message', AsyncMessage)],
            'http://aws.com', callback=self.callback,
            parent=queue,
            protocol_params={
                'json': {'AttributeNames': ['ApproximateReceiveCount']},
                'query': {'AttributeName.1': 'ApproximateReceiveCount'},
            },
        )

    def test_receive_message__with_attributes(self):
        queue = Mock(name='queue')
        self.x.receive_message(
            queue,
            self.x.get_queue_url('queue'),
            4,
            attributes=['foo', 'bar'],
            callback=self.callback,
        )
        self.x.get_list.assert_called_with(
            'ReceiveMessage', {
                'MaxNumberOfMessages': 4,
            },
            [('Message', AsyncMessage)],
            'http://aws.com', callback=self.callback,
            parent=queue,
            protocol_params={
                'json': {'AttributeNames': ['foo', 'bar']},
                'query': {'AttributeName.1': 'foo', 'AttributeName.2': 'bar'},
            },
        )

    def test_receive_message__with_fetch_attributes(self):
        queue = Mock(name='queue')
        self.x.fetch_message_attributes = ["DifferentAttribute1", "Another2"]
        self.x.receive_message(
            queue,
            self.x.get_queue_url('queue'),
            4,
            callback=self.callback,
        )
        self.x.get_list.assert_called_with(
            'ReceiveMessage', {
                'MaxNumberOfMessages': 4,
            },
            [('Message', AsyncMessage)],
            'http://aws.com', callback=self.callback,
            parent=queue,
            protocol_params={
                'json': {'AttributeNames': ['DifferentAttribute1', 'Another2']},
                'query': {'AttributeName.1': 'DifferentAttribute1', 'AttributeName.2': 'Another2'},
            },
        )

    def MockMessage(self, id=None, receipt_handle=None, body=None):
        m = Mock(name='message')
        m.id = id or uuid()
        m.receipt_handle = receipt_handle or uuid()
        m._body = body

        def _get_body():
            return m._body
        m.get_body.side_effect = _get_body

        def _set_body(value):
            m._body = value
        m.set_body.side_effect = _set_body

        return m

    def test_delete_message(self):
        queue = Mock(name='queue')
        message = self.MockMessage()
        self.x.delete_message(queue, message.receipt_handle,
                              callback=self.callback)
        self.x.get_status.assert_called_with(
            'DeleteMessage', {'ReceiptHandle': message.receipt_handle},
            queue, callback=self.callback,
        )

    def test_delete_message_batch(self):
        queue = Mock(name='queue')
        messages = [self.MockMessage('1', 'r1'),
                    self.MockMessage('2', 'r2')]
        self.x.delete_message_batch(queue, messages, callback=self.callback)
        self.x.get_object.assert_called_with(
            'DeleteMessageBatch', {},
            queue.id, verb='POST', callback=self.callback,
            protocol_params={
                'json': {'Entries': [{'Id': '1', 'ReceiptHandle': 'r1'}, {'Id': '2', 'ReceiptHandle': 'r2'}]},
                'query': {
                    'DeleteMessageBatchRequestEntry.1.Id': '1',
                    'DeleteMessageBatchRequestEntry.1.ReceiptHandle': 'r1',
                    'DeleteMessageBatchRequestEntry.2.Id': '2',
                    'DeleteMessageBatchRequestEntry.2.ReceiptHandle': 'r2',
                },
            },
        )

    def test_send_message(self):
        queue = Mock(name='queue')
        self.x.send_message(queue, 'hello', callback=self.callback)
        self.x.get_object.assert_called_with(
            'SendMessage', {'MessageBody': 'hello'},
            queue.id, verb='POST', callback=self.callback,
        )

    def test_send_message__with_delay_seconds(self):
        queue = Mock(name='queue')
        self.x.send_message(
            queue, 'hello', delay_seconds='303', callback=self.callback,
        )
        self.x.get_object.assert_called_with(
            'SendMessage', {'MessageBody': 'hello', 'DelaySeconds': 303},
            queue.id, verb='POST', callback=self.callback,
        )

    def test_send_message_batch(self):
        queue = Mock(name='queue')
        messages = [self.MockMessage('1', 'r1', 'A'),
                    self.MockMessage('2', 'r2', 'B')]
        self.x.send_message_batch(
            queue, [(m.id, m.get_body(), 303) for m in messages],
            callback=self.callback
        )
        self.x.get_object.assert_called_with(
            'SendMessageBatch', {
                'SendMessageBatchRequestEntry.1.Id': '1',
                'SendMessageBatchRequestEntry.1.MessageBody': 'A',
                'SendMessageBatchRequestEntry.1.DelaySeconds': 303,
                'SendMessageBatchRequestEntry.2.Id': '2',
                'SendMessageBatchRequestEntry.2.MessageBody': 'B',
                'SendMessageBatchRequestEntry.2.DelaySeconds': 303,
            },
            queue.id, verb='POST', callback=self.callback,
        )

    def test_change_message_visibility(self):
        queue = Mock(name='queue')
        self.x.change_message_visibility(
            queue, 'rcpt', 33, callback=self.callback,
        )
        self.x.get_status.assert_called_with(
            'ChangeMessageVisibility', {
                'ReceiptHandle': 'rcpt',
                'VisibilityTimeout': 33,
            },
            queue.id, callback=self.callback,
        )

    def test_change_message_visibility_batch(self):
        queue = Mock(name='queue')
        messages = [
            (self.MockMessage('1', 'r1'), 303),
            (self.MockMessage('2', 'r2'), 909),
        ]
        self.x.change_message_visibility_batch(
            queue, messages, callback=self.callback,
        )

        self.x.get_object.assert_called_once_with(
            'ChangeMessageVisibilityBatch', {},
            queue.id, verb='POST', callback=self.callback,
            protocol_params={
                'json': {
                    'Entries': [
                        {'Id': '1', 'ReceiptHandle': 'r1', 'VisibilityTimeout': 303},
                        {'Id': '2', 'ReceiptHandle': 'r2', 'VisibilityTimeout': 909},
                    ],
                },
                'query': {
                    'ChangeMessageVisibilityBatchRequestEntry.1.Id': '1',
                    'ChangeMessageVisibilityBatchRequestEntry.1.ReceiptHandle': 'r1',
                    'ChangeMessageVisibilityBatchRequestEntry.1.VisibilityTimeout': '303',
                    'ChangeMessageVisibilityBatchRequestEntry.2.Id': '2',
                    'ChangeMessageVisibilityBatchRequestEntry.2.ReceiptHandle': 'r2',
                    'ChangeMessageVisibilityBatchRequestEntry.2.VisibilityTimeout': '909',
                },
            },
        )

    def test_get_all_queues(self):
        self.x.get_all_queues(callback=self.callback)
        self.x.get_list.assert_called_with(
            'ListQueues', {}, [('QueueUrl', AsyncQueue)],
            callback=self.callback,
        )

    def test_get_all_queues__with_prefix(self):
        self.x.get_all_queues(prefix='kombu.', callback=self.callback)
        self.x.get_list.assert_called_with(
            'ListQueues', {'QueueNamePrefix': 'kombu.'},
            [('QueueUrl', AsyncQueue)],
            callback=self.callback,
        )

    def MockQueue(self, url):
        q = Mock(name='Queue')
        q.url = url
        return q

    def test_get_queue(self):
        self.x.get_queue('foo', callback=self.callback)
        self.x.get_list.assert_called()
        on_ready = self.x.get_list.call_args[1]['callback']
        queues = [
            self.MockQueue('/queues/bar'),
            self.MockQueue('/queues/baz'),
            self.MockQueue('/queues/foo'),
        ]
        on_ready(queues)
        self.callback.assert_called_with(queues[-1])

        self.x.get_list.assert_called_with(
            'ListQueues', {'QueueNamePrefix': 'foo'},
            [('QueueUrl', AsyncQueue)],
            callback=on_ready,
        )

    def test_get_dead_letter_source_queues(self):
        queue = Mock(name='queue')
        self.x.get_dead_letter_source_queues(queue, callback=self.callback)
        self.x.get_list.assert_called_with(
            'ListDeadLetterSourceQueues', {'QueueUrl': queue.url},
            [('QueueUrl', AsyncQueue)], callback=self.callback,
        )

    def test_add_permission(self):
        queue = Mock(name='queue')
        self.x.add_permission(
            queue, 'label', 'accid', 'action', callback=self.callback,
        )
        self.x.get_status.assert_called_with(
            'AddPermission', {
                'Label': 'label',
                'AWSAccountId': 'accid',
                'ActionName': 'action',
            },
            queue.id, callback=self.callback,
        )

    def test_remove_permission(self):
        queue = Mock(name='queue')
        self.x.remove_permission(queue, 'label', callback=self.callback)
        self.x.get_status.assert_called_with(
            'RemovePermission', {'Label': 'label'}, queue.id,
            callback=self.callback,
        )

    def test_query_protocol_encoding(self):
        assert _query_object_encode({}) == {}

        assert _query_object_encode({'Simple': 'String'}) == {'Simple': 'String'}

        assert _query_object_encode({'NumbersToString': 123}) == {'NumbersToString': '123'}

        assert _query_object_encode({'AttributeName': ['A', 'B']}) == {
            'AttributeName.1': 'A',
            'AttributeName.2': 'B',
        }

        assert _query_object_encode({'Grandparent': [{'Parent': {'Child': '1', 'Sibling': 2}}]}) == {
            'Grandparent.1.Parent.Child': '1',
            'Grandparent.1.Parent.Sibling': '2',
        }



from __future__ import annotations

from io import BytesIO
from unittest.mock import Mock, patch

import pytest
import urllib3

import t.skip
from kombu.asynchronous.http.urllib3_client import (Urllib3Client,
                                                    _get_pool_key_parts)


@t.skip.if_pypy
@pytest.mark.usefixtures('hub')
class test_Urllib3Client:
    class Client(Urllib3Client):
        urllib3 = Mock(name='urllib3')

    def test_max_clients_set(self):
        x = self.Client(max_clients=303)
        assert x.max_clients == 303

    def test_init(self):
        x = self.Client()
        assert x._pools is not None
        assert x._pending is not None
        assert x._timeout_check_tref

    def test_close(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.PoolManager'
        ):
            x = self.Client()
            x._timeout_check_tref = Mock(name='timeout_check_tref')
            x.close()
            x._timeout_check_tref.cancel.assert_called_with()
            for pool in x._pools.values():
                pool.close.assert_called_with()

    def test_add_request(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.PoolManager'
        ):
            x = self.Client()
            x._process_queue = Mock(name='_process_queue')
            request = Mock(name='request')
            x.add_request(request)
            assert request in x._pending
            x._process_queue.assert_called_with()

    def test_timeout_check(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.PoolManager'
        ):
            hub = Mock(name='hub')
            x = self.Client(hub)
            x._process_pending_requests = Mock(name='process_pending')
            x._timeout_check()
            x._process_pending_requests.assert_called_with()

    def test_process_request(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.PoolManager'
        ) as _pool_manager:
            x = self.Client()
            request = Mock(
                name='request',
                method='GET',
                url='http://example.com',
                headers={},
                body=None,
                follow_redirects=True,
                auth_username=None,
                auth_password=None,
                user_agent=None,
                use_gzip=False,
                network_interface=None,
                validate_cert=True,
                ca_certs=None,
                client_cert=None,
                client_key=None,
                proxy_host=None,
                proxy_port=None,
                proxy_username=None,
                proxy_password=None,
                on_ready=Mock(name='on_ready')
            )
            response = Mock(
                name='response',
                status=200,
                headers={},
                data=b'content'
            )
            response.geturl.return_value = 'http://example.com'
            _pool_manager.return_value.request.return_value = response

            x._process_request(request)
            response_obj = x.Response(
                request=request,
                code=200,
                headers={},
                buffer=BytesIO(b'content'),
                effective_url='http://example.com',
                error=None
            )
            request.on_ready.assert_called()
            called_response = request.on_ready.call_args[0][0]
            assert called_response.code == response_obj.code
            assert called_response.headers == response_obj.headers
            assert (
                    called_response.buffer.getvalue() ==
                    response_obj.buffer.getvalue()
            )
            assert called_response.effective_url == response_obj.effective_url
            assert called_response.error == response_obj.error

    def test_process_request_with_error(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.PoolManager'
        ) as _pool_manager:
            x = self.Client()
            x.close()
            request = Mock(
                name='request',
                method='GET',
                url='http://example.com',
                headers={},
                body=None,
                follow_redirects=True,
                auth_username=None,
                auth_password=None,
                user_agent=None,
                use_gzip=False,
                network_interface=None,
                validate_cert=True,
                ca_certs=None,
                client_cert=None,
                client_key=None,
                proxy_host=None,
                proxy_port=None,
                proxy_username=None,
                proxy_password=None,
                on_ready=Mock(name='on_ready')
            )
            _pool_manager.return_value.request.side_effect = urllib3.exceptions.HTTPError("Test Error")

            x._process_request(request)
            request.on_ready.assert_called()
            called_response = request.on_ready.call_args[0][0]
            assert called_response.code == 599
            assert called_response.error is not None
            assert called_response.error.message == "Test Error"

    def test_on_readable_on_writable(self):
        x = self.Client()
        x.on_readable(Mock(name='fd'))
        x.on_writable(Mock(name='fd'))

    def test_get_pool_with_proxy(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.ProxyManager'
        ) as _proxy_manager:
            x = self.Client()
            request = Mock(
                name='request',
                proxy_host='proxy.example.com',
                proxy_port=8080,
                proxy_username='user',
                proxy_password='pass'
            )
            x.get_pool(request)
            _proxy_manager.assert_called_with(
                proxy_url='proxy.example.com:8080',
                num_pools=x.max_clients,
                proxy_headers=urllib3.make_headers(
                    proxy_basic_auth="user:pass"
                )
            )

    def test_get_pool_without_proxy(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.PoolManager'
        ) as _pool_manager:
            x = self.Client()
            request = Mock(name='request', proxy_host=None)
            x.get_pool(request)
            _pool_manager.assert_called_with(num_pools=x.max_clients)

    def test_process_request_with_proxy(self):
        with patch(
                'kombu.asynchronous.http.urllib3_client.urllib3.ProxyManager'
        ) as _proxy_manager:
            x = self.Client()
            request = Mock(
                name='request',
                method='GET',
                url='http://example.com',
                headers={},
                body=None,
                follow_redirects=True,
                proxy_host='proxy.example.com',
                proxy_port=8080,
                proxy_username='user',
                proxy_password='pass',
                on_ready=Mock(name='on_ready')
            )
            response = Mock(
                name='response',
                status=200,
                headers={},
                data=b'content'
            )
            response.geturl.return_value = 'http://example.com'
            _proxy_manager.return_value.request.return_value = response

            x._process_request(request)
            response_obj = x.Response(
                request=request,
                code=200,
                headers={},
                buffer=BytesIO(b'content'),
                effective_url='http://example.com',
                error=None
            )
            request.on_ready.assert_called()
            called_response = request.on_ready.call_args[0][0]
            assert called_response.code == response_obj.code
            assert called_response.headers == response_obj.headers
            assert (
                    called_response.buffer.getvalue()
                    == response_obj.buffer.getvalue()
            )
            assert called_response.effective_url == response_obj.effective_url
            assert called_response.error == response_obj.error

    def test_pool_key_parts(self):
        request = Mock(
            name='request',
            method='GET',
            url='http://example.com',
            headers={},
            body=None,
            network_interface='test',
            validate_cert=False,
            ca_certs='test0.pem',
            client_cert='test1.pem',
            client_key='some_key',
        )
        pool_key = _get_pool_key_parts(request)
        assert pool_key == [
            "interface=test",
            "validate_cert=False",
            "ca_certs=test0.pem",
            "client_cert=test1.pem",
            "client_key=some_key"
        ]



from __future__ import annotations

import pytest

import t.skip

pytest.importorskip('boto3')
pytest.importorskip('urllib3')


@t.skip.if_pypy
@pytest.mark.usefixtures('hub')
class AWSCase:
    pass
